/*
 * Copyright (C) 2000, iAgora LLC.
 * Written by Roger Espel Llima for iAgora LLC, 28-NOV-2000.
 *
 * This is the mighty linger daemon, a devilish contraption that
 * specializes in doing nothing to network sockets, which, despite all
 * appearances, is actually a useful thing to do.  This saves Apache
 * from doing this boring work, and does it more effectively than Apache
 * could, by handling thousands of linger operations within a single
 * thread, and proving once again that select() (err, I mean, poll())
 * rules the world of networking.
 *
 * This program is Free Software.  You can redistribute and/or modify it
 * under the terms specified in the file ``LICENSE'', which is part of
 * the lingerd distribution.  (For the impatient: it's an Apache-style
 * license, without the advertising clause.)
 *
 */

#define IN_LINGERD
#include "lingerd.h"
#include "li_config.h"

#undef DEBUG

#define VERSION "0.94"

/**********************************************************************
 *                     Global data starts here                        *
 **********************************************************************/

/* 
 * Data on all open connections is held in an array of fd_info's
 * and an array of poll structs, indexed in the same sequence.
 */
fd_info		*all_fds   = NULL;
poll_entry	*polls     = NULL;

/*
 * Max number of sockets; this is the size of the fd_info and poll_info 
 * arrays, and doesn't change after initialization.
 */
int	max_socks   = 0;

/* Number of currently open sockets */
int	n_sockets   = 0;

/* Number of currently open sockets that correspond to lingering fds */
int	linger_fds  = 0;

/* The current time */
struct timeval	now;

/* 
 * Restart scan flag: set by evict_fd() when the order of the fd_info
 * array changes, in which case the scan loop must restart from the 
 * beginning.
 */
int	restart_scan = 0;

/*
 * The nodaemon flag; set to 1 if we're not a daemon (in which case we
 * log to stderr, not syslog).
 */
int	no_daemon   = 0;

/*
 * The no_stats flag; if set, we don't generate hourly (or whatever) stats.
 */
int 	no_stats    = 0;

/* Path to the Unix-domain socket */
char	*sock_path  = SOCKPATH;

/* Path to the pid-file */
char	*pid_file   = PID_FILE;

/* 
 * Last modification time of the pid file and of the Unix-domain socket
 * (only delete them at cleanup time it if they haven't changed).
 */
time_t	pid_lastmod  = (time_t)-1;
time_t	sock_lastmod = (time_t)-1;

/* 
 * Some global counters for statistics.  These count the total number of
 * clients and linger sockets handled since the last statistics print;
 * they don't decrease on connection end.
 */
int 	n_lingers   = 0;
int	n_clients   = 0;
int	n_errors    = 0;


/************************* End of global data ***************************/

/*
 * Returns the mtime (last modification) of a file, or -1 in case of
 * failure.
 */
time_t mtime(char *fn) {
  struct stat st;

  if (stat(fn, &st) == 0)
    return st.st_mtime;
  
  return (time_t)-1;
}

/*
 * The main logging function, works like syslog() but outputs to stderr
 * if we're not a daemon.
 */
void log(int priority, char *fmt, ...) {
  va_list ap;
  char buf[256];

  va_start(ap, fmt);

  if (no_daemon) {
    fprintf(stderr, "%s: ", MY_NAME);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
  } else {
    vsnprintf(buf, sizeof(buf), fmt, ap);
    buf[255] = 0; /* just in case */
    syslog(priority, "%s", buf);
  }

  va_end(ap);
}

/* 
 * This function works like perror(), but logs to syslog if appropriate,
 * and exits the program.
 */
void pdie(char *s) {
  log(LOG_CRIT, "%s: %s", s, strerror(errno));
  cleanup_exit(1);
}

/*
 * Exit, cleaning up the socket and the pid file (unless they've been 
 * changed since we last touched them.
 */
void cleanup_exit(int err_code) {

  /* Remove the listening socket */
  if (sock_lastmod > 0 && sock_lastmod == mtime(sock_path))
    unlink(sock_path);
  
  /* Remove the pid file */
  if (!no_daemon && pid_file != NULL && 
      pid_lastmod > 0 && pid_lastmod == mtime(pid_file))
    unlink(pid_file);

  exit(err_code);
}

void signal_exit(int sig) {
  if (!no_stats)
    log(LOG_NOTICE, "Caught signal %d - exiting", sig);
  cleanup_exit(0);
}

/* 
 * Set up signal handling; ignore HUP and PIPE, and die gracefully
 * on TERM and INT.
 */
void init_signals(void) {
  signal(SIGHUP,  SIG_IGN);
  signal(SIGPIPE, SIG_IGN);
  signal(SIGTERM, signal_exit);
  signal(SIGINT,  signal_exit);
}

/* 
 * Become a well-behaved daemon.  Call this before init_logging(), or it
 * might close the syslog fd, and call it before writing the pid because
 * forking changes it.
 */
void go_daemon(void) {
  int i;

  /* Close all reasonable fd's; set stdin/out/err to /dev/null */
  for (i=0; i<256; i++)
    close(i);
  open("/dev/null", O_RDONLY);
  open("/dev/null", O_WRONLY);
  open("/dev/null", O_WRONLY);

  /* Detach from the current directory and from the controlling terminal */
  chdir("/");
  if (fork()) _exit(0);
  setsid();
  if (fork()) _exit(0);
}

/*
 * Read the pid value from the pid file; returns -1 on failure.
 */
int read_pidfile(void) {
  char buf[16];
  int fd = -1, n, pid;

  if (pid_file == NULL)
    return -1;

  if ((fd = open(pid_file, O_RDONLY)) < 0 ||
      (n  = read(fd, buf, sizeof(buf))) <= 0) {
    if (fd > 0) close(fd);
    return -1;
  }

  close(fd);
  buf[15] = 0;
  pid = atoi(buf);

  return (pid > 0 ? pid : -1);
}

/* 
 * Write our own pid to the specified pidfile.  Check and log if there
 * was already a lingerd running.
 */
void write_pidfile(void) {
  char	buf[16];
  int	pid, fd, n;

  if (pid_file == NULL)
    return;

  if ((pid = read_pidfile()) > 1 && kill(pid, 0) == 0) {
    log(LOG_WARNING, "There is already a lingerd running with pid %d;", pid);
    log(LOG_WARNING, "you might want to manually kill it.");
  }

  if ((fd = open(pid_file, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0)
    return;

  n = snprintf(buf, sizeof(buf), "%d\n", getpid());
  if (n > 0) write(fd, buf, n);
  close(fd);

  /* Keep track of the last-modified time of the pid file */
  pid_lastmod = mtime(pid_file);
}

/* 
 * Read the pid file and kill a running lingerd.
 */
void kill_running_lingerd(void) {
  int pid;

  if (pid_file == NULL) {
    fprintf(stderr, "PID_FILE not defined!\n");
    exit(1);
  }

  pid = read_pidfile();
  if (pid <= 0) {
    fprintf(stderr, "Can't read pid file (%s).\n", pid_file);
    exit(1);
  }

  if (kill(pid, SIGTERM) < 0) {
    fprintf(stderr, "Can't send SIGTERM to process %d.", pid);
    exit(1);
  }

  exit(0);
}

/* 
 * Make a socket non-blocking.  Returns 0 on success, -1 on failure.
 */
int make_nonblocking(int fd) {
  int r;
  
  if ((r = fcntl(fd, F_GETFL, 0)) < 0 ||
      fcntl(fd, F_SETFL, r|O_NONBLOCK) < 0)
    return -1;

  return 0;
}

/* 
 * This function gets a passed fd from a Unix-domain socket.  If it succeeds,
 * it returns the fd; if there's an EOF on the socket, it returns -1.  If
 * there's any other error, it returns a different negative number < -1 
 * for each possible error.
 */

int recv_fd(int sockfd) {
#ifndef OLD_FD_PASSING
  lingerd_cmsghdr l_hdr;
  struct cmsghdr *hdr = &l_hdr.hdr;
#endif
  struct msghdr msg;
  struct iovec	iov;
  int fd, r;
  /* 
   * We're not really expecting much in terms of ordiary socket data,
   * but if someone's silly enough to send us lots of it, might as well
   * read it in decent-sized amounts.
   */
  char buf[4096];

#ifndef OLD_FD_PASSING
  memset((char *)&l_hdr, 0, sizeof(l_hdr));
#endif
  memset((char *)&msg,   0, sizeof(msg));

  iov.iov_base    = buf;
  iov.iov_len     = sizeof(buf);
  msg.msg_iov     = &iov;
  msg.msg_iovlen  = 1;
  msg.msg_name    = NULL;
  msg.msg_namelen = 0;

#ifdef OLD_FD_PASSING
  /* log(LOG_DEBUG, "recv_fd : using old fd passing"); */
  msg.msg_accrights = (char *)&fd;
  msg.msg_accrightslen = sizeof(fd);
  fd = -1;
#else
  msg.msg_flags	  = 0;
  msg.msg_control = (char *)hdr;
  msg.msg_controllen = CONTROLLEN;
  *(int *)CMSG_DATA(hdr) = -1;
  /* log(LOG_DEBUG, "recv_fd : using new fd passing; controllen = %d", CONTROLLEN); */
#endif

  if ((r = recvmsg(sockfd, &msg, 0)) == 0)
    return -1; /* EOF */

  /* log(LOG_DEBUG, "recv_fd : recvmsg returned %d", r); */

  /* Respond to a primitive ping? -> pong! protocol */
  if (r >= 5 && !strncmp(buf, "ping?", 5))
    (void)write(sockfd, "pong!", 6);

#ifdef OLD_FD_PASSING
  if (r < 0 || msg.msg_accrightslen != sizeof(fd) || fd < 0) {
    /* Got something but it's not a proper fd */
    /* log(LOG_DEBUG, "recv_fd : got something wrong (accrightslen = %d, fd = %d)", msg.msg_accrightslen, fd); */
    return -2;
  }
#else
  fd = *(int *)CMSG_DATA(hdr);
  if (r < 0 || fd < 0 || 
      msg.msg_controllen != CONTROLLEN ||
      hdr->cmsg_type     != SCM_RIGHTS ||
      hdr->cmsg_level    != SOL_SOCKET ||
      hdr->cmsg_len      != CONTROLLEN) {
    /* log(LOG_DEBUG, "recv_fd : got something wrong (fd = %d, msg_controllen = %d, hdr->cmsg_type = %d, hdr->cmsg_level = %d, hdr->cmsg_len = %d)", fd, msg.msg_controllen, hdr->cmsg_type, hdr->cmsg_level, hdr->cmsg_len); */
    /* Got something but it's not a proper fd */
    return -2;
  }
#endif

  /* log(LOG_DEBUG, "recv_fd : got fd = %d", fd); */

  /* We have an fd; make sure it's a socket */

#if 0
/* getpeername() fails on fd-passed sockets on Solaris, go figure why */
  {
    struct sockaddr_in sin;
    int len = sizeof(sin);

    if ((r = getpeername(fd, (struct sockaddr *)&sin, &len)) < 0) {
      log(LOG_NOTICE, "recv_fd: getpeername() failed");
      close(fd);
      return -3;
    }
  }
#else
  {
    int opt = 0, optlen = sizeof(int);

    if ((r = getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &opt, &optlen)) < 0) {
      /* log(LOG_DEBUG, "recv_fd: getsockopt() failed"); */
      close(fd);
      return -3;
    }
  }
#endif

  return fd;
}

/*
 * Sleep for 50ms; called every time there's an unexpected error,
 * to avoid tight loops in case something goes haywire.
 */
void sleepabit(void) {
  struct timeval tv;
  tv.tv_sec = 0;
  tv.tv_usec = 50000;
  select(1, NULL, NULL, NULL, &tv);
}

/* 
 * Ask the system for the maximum number of open fd's in this process;
 * move it up if possible, and return the new maximum.
 */
int get_max_fds(void) {
  struct rlimit rlim;
  int r;

  /* Get max number of open fd's */
  r = getrlimit(RLIMIT_NOFILE, &rlim);
  if (r < 0)
    pdie("getrlimit() failed");
  
  /* Set it up if needed */
  if (rlim.rlim_cur < rlim.rlim_max) {
    rlim.rlim_cur = rlim.rlim_max;
    setrlimit(RLIMIT_NOFILE, &rlim);
    /* Check again just in case */
    r = getrlimit(RLIMIT_NOFILE, &rlim);
    if (r < 0)
      pdie("getrlimit() failed");
  }

  /* Make sure the fd limit is within the expected bounds */
  r = (rlim.rlim_cur <= MAX_FD_LIMIT ? rlim.rlim_cur : MAX_FD_LIMIT);

  if (r < MIN_FD_LIMIT) {
    log(LOG_CRIT, "Can only get %d sockets; I want %d",
	r, MIN_FD_LIMIT);
    cleanup_exit(1);
  }

  if (r <= SOCK_SETASIDE + 1) {
    log(LOG_CRIT, "Can only get %d sockets; I want at least %d", r, 
	  2 + SOCK_SETASIDE);
    cleanup_exit(1);
  }

  return r;
}

void init_logging(void) {
  openlog(MY_NAME, LOG_PID, LOG_FACILITY);
}

/*
 * Initialize the global data structures; figure out the max number of
 * fd's and allocate the arrays.
 */
void init_data(void) {
  max_socks = get_max_fds() - SOCK_SETASIDE;
  n_sockets = 0;
  all_fds   = (fd_info    *)malloc(max_socks * sizeof(fd_info));
  polls     = (poll_entry *)malloc(max_socks * sizeof(poll_entry));

  if (all_fds == NULL || polls == NULL)
    pdie("malloc() failed");
}

/*
 * Remove an fd from the global array, closing it in the process.  The
 * parameter n is not the fd, but its index in the array.
 */
void evict_fd(int n) {
  int fd   = all_fds[n].fd;

  /* If we're closing the listening socket, then we're gone. */
  if (all_fds[n].callback == listen_callback) {
    log(LOG_CRIT, "evict_fd: closing the listening socket!");
    cleanup_exit(1);
  }

  /* Keep track of the number of lingering sockets */
  if (all_fds[n].callback == linger_callback)
    linger_fds--;

  close(fd);
  n_sockets--;

  /* Move the last entry of the arrays over the dead one */
  if (n < n_sockets) {
    all_fds[n] = all_fds[n_sockets];
    polls  [n] = polls  [n_sockets];

    /* 
     * Tell the scan loop to restart from the beginning; otherwise
     * it would miss the socket we just moved.
     */
    restart_scan = 1;
  }

  if (n_sockets <= 0) {
    log(LOG_CRIT, "evict_fd: n_sockets = %d, should be > 0 always!");
    cleanup_exit(1);
  }
}

/* 
 * Try hard to free a socket, by closing the oldest lingering connection
 * in the array.
 */
void free_a_socket(void) {
  int	i, oldest = -1;
  time_t oldtime  = 1 + NOW();

  for (i=0; i<n_sockets; i++) {
    if (all_fds[i].callback == linger_callback &&
	all_fds[i].insert_time < oldtime) {
      oldtime = all_fds[i].insert_time;
      oldest  = i;
    }
  }
  
  if (oldest > 0)
    evict_fd(oldest);
}

/*
 * Add an fd to the global array.
 */
void add_fd(fd_info *fdi) {
  poll_entry pe;

  /* If we're out of sockets, try to free one */
  if (n_sockets >= max_socks)
    free_a_socket();

  /* 
   * If we're *really* out of sockets, we'll just have to close the
   * incoming one, and log it.
   */
  if (n_sockets >= max_socks) {
    log(LOG_ERR, "add_fd: out of sockets! not adding fd (%d)", fdi->fd);
    n_errors++;
    close(fdi->fd);
    return;
  }

  fdi->insert_time = NOW();

  /* 
   * Set up the poll_entry, and put it in the array, and set up the
   * fd_info.
   */
  pe.fd	     = fdi->fd;
  pe.events  = POLLIN;
  pe.revents = 0;

  all_fds[n_sockets] = *fdi;
  polls  [n_sockets] = pe;
  n_sockets++;

  /* 
   * Keep track of the number of lingering sockets; only increment it
   * now that we're sure this one has been added.
   */
  if (fdi->callback == linger_callback)
    linger_fds++;
}

/*
 * Check the expiration status of a lingering socket; kill it if it's
 * past its expiration date (soft or hard).  The argument n is the index
 * in the all_fds array.  This gets called a lot so let's try to make it
 * inline (a GCC extension).
 */

#ifdef __GNUC__
inline
#endif
void check_expire(int n) {
  register fd_info *fdi = &all_fds[n];

  if ((fdi->linger_time && NOW() >= fdi->linger_time) ||
      (fdi->linger_max  && NOW() >= fdi->linger_max))
    evict_fd(n);
}


/*
 * This is the main linger callback, which gets called whenever there's
 * something to read from a lingering socket.  This means either data or
 * EOF.  Like all callbacks, this routine is not allowed to block.
 */
void linger_callback(int n, int fd) {
  char	buf[4096];
  int	total = 0, r;
  fd_info *fdi = &all_fds[n];

  do {
    /* Read from the socket, see what we get */
    r = read(fd, buf, sizeof(buf));
    
    /* 
     * EAGAIN is not really an error, it just means there's nothing
     * more to read.
     */
    if (r == -1 && (errno == EAGAIN || errno == EWOULDBLOCK))
      r = 0;

    if (r < 0 || (r == 0 && total == 0)) {
      /* If we got either an EOF or an error, kill the socket. */
      evict_fd(n);
      return;
    }

    total += r;
  
    /* 
     * Keep reading as long as we got a full buffer (if we got less
     * than that it means there's nothing more to read w/o blocking).
     * Stop at 16 iterations because we don't want a single client to
     * stall us for too long.
     */
  } while (r == sizeof(buf) && total < 16*sizeof(buf));

  /* Update the soft deadline if we managed to read something */
  if (total > 0)
    fdi->linger_time = NOW() + LINGER_TIME;

  /* Check expiration status */
  check_expire(n);
}

/*
 * This callback gets called whenever there's something to read from
 * a client socket.  This means either EOF or a passed fd.
 */
void client_callback(int n, int client_fd) {
  int	getfd;
  fd_info	fdi;

  getfd = recv_fd(client_fd);

  /* EOF on the client fd; evict it and we're done.  */
  if (getfd == -1) {
    evict_fd(n);
    return;
  }

  /* recv_fd() failure */
  if (getfd < 0) {
    /* -3 is not really an error, it just means the other side closed the
     * socket before we got a chance to look at it.
     */
    if (getfd != -3) {
      log(LOG_WARNING, "recv_fd failed (%d)!", getfd);
      n_errors++;
    }
    return;
  }

  /* Make the socket non-blocking; kill it if that fails */
  if (make_nonblocking(getfd) < 0) {
    log(LOG_WARNING, "client_callback: Can't make socket nonblocking!");
    n_errors++;
    close(getfd);
    return;
  }

  /* 
   * No shutdown business; this is done on the Apache side.  Otherwise
   * we end up with two fd's on the same socket, one of them doing a
   * shutdown and the other a close, in an undetermined order, and I
   * don't want to know what the kernel would do in that case.
   */

#if 0
  /* Shut down the writing side of the socket, i.e send the client a
   * FIN.
   */
  if (shutdown(getfd, 1) < 0) {
    log(LOG_WARNING, "client_callback: Can't shutdown socket for writing");
    n_errors++;
    close(getfd);
    return;
  }
#endif

  /* Add it to the list of sockets to watch */
  fdi.fd	  = getfd;
  fdi.callback	  = linger_callback;
  fdi.linger_time = NOW() + LINGER_TIME;
  fdi.linger_max  = NOW() + LINGER_MAX;

  add_fd(&fdi);

  n_lingers++;
}

/* 
 * This function gets called whenever the listen socket is ready to read
 * (ie. to accept()).
 */
void listen_callback(int n, int listenfd) {
  struct sockaddr_un sadr;
  int	sockfd, adrlen = sizeof(sadr);
  fd_info	fdi;

  sockfd = accept(listenfd, (struct sockaddr *)&sadr, &adrlen);

  if (sockfd < 0) {
    log(LOG_WARNING, "listen_callback: accept() failed, errno = %d", errno);
    n_errors++;
    sleepabit();
    return;
  }

  /* Make the socket non-blocking; kill it if that fails */
  if (make_nonblocking(sockfd) < 0) {
    log(LOG_WARNING, "listen_callback: Can't make socket nonblocking!");
    n_errors++;
    close(sockfd);
    return;
  }

  /* Add it to the list of sockets to watch */
  fdi.fd	  = sockfd;
  fdi.callback	  = client_callback;
  fdi.linger_time = (time_t)0;
  fdi.linger_max  = (time_t)0;

  add_fd(&fdi);

  n_clients++;
}

/*
 * Set up the listening Unix-domain socket, and add it to the global
 * fd array.
 */
void init_listen_sock(void) {
  int		fd;
  fd_info	fdi;
  struct sockaddr_un sadr;

  /* 
   * Clear the socket from the filesystem, in case it was there from a
   * previous run.  This call is allowed to fail.
   */
  unlink(sock_path);

  if ((fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
    pdie("socket() failed");

  sadr.sun_family = AF_UNIX;
  strncpy(sadr.sun_path, sock_path, sizeof(sadr.sun_path));
  sadr.sun_path[sizeof(sadr.sun_path) - 1] = '\0';

  if (bind(fd, (struct sockaddr *)&sadr, sizeof(sadr)) < 0)
    pdie("bind() failed");

  sock_lastmod = mtime(sock_path);

  if (listen(fd, BACKLOG) < 0)
    pdie("listen() failed");

  fdi.fd	  = fd;
  fdi.callback	  = listen_callback;
  fdi.linger_time = (time_t)0;
  fdi.linger_max  = (time_t)0;

  add_fd(&fdi);
}

/* 
 * Scan the array of poll() return values, running callbacks on sockets
 * with activity, and checking expiration on the others.
 */
void scan_socks(void) {
  int	i = 0, rev, iterations = 0, nsocks;

  /* 
   * Reset the restart flag; it will be set by evict_fd if we need to
   * start scanning again from the beginning of the array.
   */
  restart_scan = 0;

  /*
   * Save a working copy of n_sockets, so that the scan loop doesn't
   * reach any new sockets that might get added by callbacks.  This is
   * not perfect because of evictions, but that's OK because new
   * sockets get inserted with pe.revents = 0 anyway, so they won't
   * trigger any callbacks.  But we need to check against n_sockets
   * too, in case it has decreased in the process.
   */
  nsocks = n_sockets;

  while (i < nsocks && i < n_sockets) {
    /*
     * Get the status of the socket; kill it if there's an error or an
     * EOF, call the callback if there's something to read, and check
     * expiration otherwise.  Make sure to clear the status so we don't
     * handle the socket twice during the same scan run.
     */
    rev = polls[i].revents;
    polls[i].revents = 0;
    if (rev & (POLLERR | POLLHUP | POLLNVAL)) {
      /* Error => kill the socket */
      evict_fd(i);
    } else if (rev & POLLIN) {
      /* Ready for reading => call the callback */
      (all_fds[i].callback)(i, polls[i].fd);
    } else {
      /* None of the above => just check for expiration */
      check_expire(i);
    }

    i++;
    if (restart_scan)
      i = restart_scan = 0;

    if (++iterations > 100000) {
      /* Damn, did we just get stuck in an endless loop?! */
      log(LOG_WARNING, "scan_socks: too many iterations, leaving loop");
      n_errors++;
      return;
    }
  }
}

/* 
 * Print some (rather useless) statistics about the activity of the linger
 * daemon.
 */
void do_stats(void) {
  static int last_time     = 0;
  static int total_lingers = 0; 
  static int total_clients = 0;
  static int total_errors  = 0;

  if (NOW() < last_time + STAT_PERIOD)
    return;

  last_time = NOW();

  total_lingers += n_lingers;
  total_clients += n_clients;
  total_errors  += n_errors;

  log(LOG_INFO,
      "Handled %d lingers, %d clients, %d errors (total: %d, %d, %d)", 
      n_lingers, n_clients, n_errors, 
      total_lingers, total_clients, total_errors);
  log(LOG_INFO,
      "I have %d open sockets (max %d), including %d lingers",
      n_sockets, max_socks, linger_fds);

  n_lingers = n_clients = n_errors = 0;
}

/* 
 * The main() routine.  Do everything.
 */
int main (int argc, char *argv[]) {
  int i;
  char *arg;

  /* Read and handle command-line arguments */
  for (i=1; i<argc; i++) {
    arg = argv[i];
    
    /* -option and --option are equivalent */
    if (arg[0] == '-' && arg[1] == '-')
      arg++;

    if (!strcmp(arg, "-X") || !strcmp(arg, "-nodaemon") || !strcmp(arg, "-d"))
      no_daemon = 1;

    else if (!strcmp(arg, "-q") || !strcmp(arg, "-quiet"))
      no_stats	= 1;

    else if (!strcmp(arg, "-k") || !strcmp(arg, "-kill"))
      kill_running_lingerd();

    else if (!strcmp(arg, "-V") || !strcmp(arg, "-version")) {
      fprintf(stderr, "%s version %s.\n", MY_NAME, VERSION);
      exit(0);

    } else if (!strcmp(arg, "-s") || !strcmp(arg, "-socket")) {
      sock_path = argv[++i];
      if (!sock_path) {
	fprintf(stderr, "%s: I want an argument for -s", MY_NAME);
	exit(1);
      }

    } else if (!strcmp(arg, "-p") || !strcmp(arg, "-pidfile")) {
      pid_file = argv[++i];
      if (!pid_file) {
	fprintf(stderr, "%s: I want an argument for -p", MY_NAME);
	exit(1);
      } else if (*pid_file != '/') {
	/* if the pidfile doesn't start with a /, treat it as "no pidfile" */
        pid_file = NULL;
      }

    } else if (!strcmp(arg, "-h") || !strcmp(arg, "-help")) {
      fprintf(stderr,

"This is lingerd version " VERSION ", the mighty lingering daemon.\n\n"
"Use: lingerd [options]\n"
"Options: \n"
"     -X, -nodaemon, -d   --  run from the terminal, with errors to\n"
"                             stderr instead of syslog\n"
"     -q, -quiet          --  do not generate periodic statistics\n"
"     -k, -kill           --  kill a running lingerd\n"
"     -V, -version        --  show version number\n"
"     -s, -socket [path]  --  specify the path (filename) of the\n"
"                             Unix-domain socket\n"
"     -p, -pidfile [path] --  specify the path (filename) of the pid file\n\n"

      );
      exit(0);

    } else {
      fprintf(stderr, "%s: unrecognized option: `%s'; try %s --help\n", MY_NAME,
	      arg, MY_NAME);
      exit(1);
    }
  }

  /* Initialize everything, in the right order */

  if (!no_daemon)
    go_daemon();

  init_logging();

  if (!no_daemon)
    write_pidfile();

  init_data();
  init_signals();
  init_listen_sock();

  if (!no_stats)
    log(LOG_NOTICE, "Linger daemon started; socket table has %d entries.",
	max_socks);

  /* The main loop */
  while (1) {
    int timeout = linger_fds > 0 ? POLL_TIMEOUT : SLOW_TIMEOUT;

    if (poll(polls, n_sockets, timeout) < 0) {
      /* 
       * Poll really shouldn't fail, because badfd errors are reported
       * in polls[].  If it does fail, complain and restart.
       */
       if (errno == EAGAIN || errno == EINTR)
	 continue;
       log(LOG_WARNING, "poll() failure (errno = %d), continuing", errno);
       n_errors++;
       sleepabit();
       continue;
    }

    gettimeofday(&now, NULL);

    /* Run the scan loop */
    scan_socks();

    /* Maybe print some stats */
    if (STAT_PERIOD > 0 && !no_stats)
      do_stats();

  }
}

